# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pprint
import sys
import time

import pulp
import pytest

from slambuc.alg import LP_LAT
from slambuc.alg.chain.ser.ilp import (build_chain_mtx_model, chain_mtx_partitioning, recreate_blocks_from_xmatrix,
                                       extract_blocks_from_xmatrix, build_greedy_chain_mtx_model)
from slambuc.misc.generator import get_random_chain_data
from slambuc.misc.util import (print_lp_desc, evaluate_ser_chain_partitioning, print_ser_chain_summary,
                               print_var_matrix, print_pulp_matrix_values, print_cost_coeffs, print_lat_coeffs,
                               get_cplex_path, get_glpk_path)


def test_mtx_model_creation(save_file: bool = False):
    params = dict(runtime=[20, 10, 30, 20],
                  memory=[3, 3, 2, 1],
                  rate=[1, 2, 2, 1],
                  data=[5, 2, 5, 3],
                  M=6,
                  L=50,
                  # start=1,
                  # end=2,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    #
    _s = time.perf_counter()
    model_greedy, X_greedy = build_greedy_chain_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"Greedy Model building time: {_d * 1000} ms")
    _s = time.perf_counter()
    model, X = build_chain_mtx_model(**params)
    _d = time.perf_counter() - _s
    #
    print(f"Direct Model building time: {_d * 1000} ms")
    print("  Decision variables  ".center(80, '='))
    print("Greedy:")
    print_var_matrix(X_greedy)
    print("Direct:")
    print_var_matrix(X)
    print("  Cost Coefficients  ".center(80, '='))
    print("Greedy:")
    print_cost_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_cost_coeffs(model, X)
    print("  Latency Coefficients  ".center(80, '='))
    print("Greedy:")
    print_lat_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_lat_coeffs(model, X)
    print("  Generated LP model  ".center(80, '='))
    print_lp_desc(model_greedy)
    print_lp_desc(model)
    if save_file:
        model_greedy.writeLP("chain_mtx_model.lp")


def test_mtx_model_solution():
    params = dict(runtime=[20, 10, 30, 20],
                  memory=[3, 3, 2, 1],
                  rate=[1, 2, 2, 1],
                  data=[5, 2, 5, 3],
                  M=6,
                  L=150,
                  # start=1,
                  # end=2,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    print(" Greedy model ".center(80, '='))
    model, X = build_greedy_chain_mtx_model(**params)
    status = model.solve(solver=pulp.PULP_CBC_CMD(mip=True, msg=True))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    #
    print(" Direct model ".center(80, '='))
    model, X = build_chain_mtx_model(**params)
    status = model.solve(solver=pulp.PULP_CBC_CMD(mip=True, msg=True))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    #
    print("Solution:")
    print_pulp_matrix_values(X)
    _s = time.perf_counter()
    rec_partition = recreate_blocks_from_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Recreate: {_d * 1000} ms")
    _s = time.perf_counter()
    ext_partition = extract_blocks_from_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {rec_partition = }, {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_mtx_model_solution_cplex():
    params = dict(runtime=[20, 10, 30, 20],
                  memory=[3, 3, 2, 1],
                  rate=[1, 2, 2, 1],
                  data=[5, 2, 5, 3],
                  M=6,
                  L=150,
                  start=1,
                  end=2,
                  delay=10)
    print("  Run CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = chain_mtx_partitioning(**params, solver=pulp.CPLEX_PY(mip=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


@pytest.mark.skipif(get_glpk_path() is None, reason="GLPK is not available!")
def test_mtx_model_solution_glpk():
    params = dict(runtime=[20, 10, 30, 20],
                  memory=[3, 3, 2, 1],
                  rate=[1, 2, 2, 1],
                  data=[5, 2, 5, 3],
                  M=6,
                  L=150,
                  start=1,
                  end=2,
                  delay=10)
    print("  Run GLPK solver  ".center(80, '='))
    partition, opt_cost, opt_lat = chain_mtx_partitioning(**params, solver=pulp.GLPK_CMD(mip=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


def evaluate_mtx_model():
    params = dict(runtime=[20, 10, 30, 20],
                  memory=[3, 3, 2, 1],
                  rate=[1, 2, 2, 1],
                  data=[5, 2, 5, 3],
                  M=6,
                  L=50,
                  start=1,
                  end=2,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = chain_mtx_partitioning(**params, solver=pulp.PULP_CBC_CMD(mip=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_chain_partitioning(partition, opt_cost, opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = chain_mtx_partitioning(**params, solver=pulp.CPLEX_PY(mip=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_chain_partitioning(partition, opt_cost, opt_lat, **params)


########################################################################################################################


def run_test(runtime: list, memory: list, rate: list, data: list, M: int = math.inf, L: int = math.inf, delay: int = 1):
    partition, opt_cost, opt_lat = chain_mtx_partitioning(runtime, memory, rate, data, M, L, delay,
                                                          solver=pulp.PULP_CBC_CMD(mip=True, msg=False))
    evaluate_ser_chain_partitioning(partition, opt_cost, opt_lat, runtime, memory, rate, data, M, L, 0, None, delay)
    return partition, opt_cost, opt_lat


def test_ser_chain():
    params = dict(runtime=[20, 40, 50, 20, 70, 40, 50, 60, 40, 10],
                  memory=[3, 3, 2, 1, 2, 1, 2, 1, 2, 3],
                  rate=[1, 1, 2, 2, 1, 3, 1, 2, 1, 3],
                  data=[5, 3, 5, 2, 1, 3, 2, 3, 5, 1],
                  delay=10,
                  M=6,
                  L=800)
    print_ser_chain_summary(params['runtime'], params['memory'], params['rate'], params['data'])
    run_test(**params)


def test_random_ser_chain(n: int = 10):
    runtime, memory, rate, data = get_random_chain_data(n, (10, 100), (1, 3), (1, 3), (1, 5))
    params = dict(runtime=runtime,
                  memory=memory,
                  rate=rate,
                  data=data,
                  delay=10,
                  M=6,
                  L=700)
    print_ser_chain_summary(runtime, memory, rate, data)
    run_test(**params)


def test_partial_ser_chain():
    params = dict(runtime=[20, 40, 50, 20, 70, 40, 50, 60, 40, 10],
                  memory=[3, 3, 2, 1, 2, 1, 2, 1, 2, 3],
                  rate=[1, 1, 2, 2, 1, 3, 1, 2, 1, 3],
                  data=[5, 3, 5, 2, 1, 3, 2, 3, 5, 1],
                  delay=10,
                  M=6,
                  L=math.inf)
    print_ser_chain_summary(params['runtime'], params['memory'], params['rate'], params['data'])
    # No restriction
    run_test(**params)
    # Optimal
    params['L'] = 702
    run_test(**params)
    # Stricter restriction
    for l in [600, 540, 535, 525, 465]:
        params['L'] = l
        run_test(**params)
    # Infeasible due to M
    params['L'] = 455
    run_test(**params)


if __name__ == '__main__':
    # test_mtx_model_creation(save_file=False)
    # test_mtx_model_solution()
    # test_mtx_model_solution_cplex()
    test_mtx_model_solution_glpk()
    # evaluate_mtx_model()
    # test_ser_chain()
    # test_random_ser_chain()
    # test_partial_ser_chain()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib

import networkx as nx

from slambuc.alg.service import NAME
from slambuc.alg.tree.ser.pseudo_mp import (isubtree_cutoffs, isubtree_sync_cutoffs, isubtree_splits, get_cpu_splits,
                                            pseudo_mp_btree_partitioning, pseudo_mp_ltree_partitioning)
from slambuc.misc.generator import get_random_tree
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import evaluate_ser_tree_partitioning


def test_cpu_cutoff(tree: nx.DiGraph = None, cut_factor: int = None, draw: bool = True):
    if not tree:
        tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo_mp"
    cut_factor = cut_factor if cut_factor else math.ceil(math.sqrt(len(tree) - 1))
    #
    bounds = dict(lb=1, ub=math.inf)
    print(f" All non-trivial cutoffs ({bounds}) ".center(80, '='))
    coffs = list(isubtree_cutoffs(tree, root=1, **bounds))
    for i, (e, c) in enumerate(coffs):
        print(f"{i}.  {e[0]} -> {e[1]} : {c}")
    bounds.update(ub=cut_factor)
    print(f" All candidate cutoffs ({bounds}) ".center(80, '='))
    coffs = list(isubtree_cutoffs(tree, root=1, **bounds))
    for i, (e, c) in enumerate(coffs):
        print(f"{i}.  {e[0]} -> {e[1]} : {c}")
    bounds.update(lb=cut_factor - 1)
    print(f" Designated cutoffs ({bounds}) ".center(80, '='))
    coffs = list(isubtree_cutoffs(tree, root=1, **bounds))
    for i, (e, c) in enumerate(coffs):
        print(f"{i}.  {e[0]} -> {e[1]} : {c}")
    bounds.update(lb=cut_factor)
    print(f" Designated cutoffs ({bounds}) ".center(80, '='))
    coffs = list(isubtree_cutoffs(tree, root=1, **bounds))
    for i, (e, c) in enumerate(coffs):
        print(f"{i}.  {e[0]} -> {e[1]} : {c}")
    if draw:
        draw_tree(tree, cuts=[e for e, _ in coffs])


def test_subtree_split(tree: nx.DiGraph = None, size: int = None, draw: bool = True):
    if not tree:
        tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo_mp"
    size = size if size else math.ceil(math.sqrt(len(tree) - 1))
    #
    bounds = dict(lb=1, ub=size)
    print(f" All candidate cutoffs ({bounds}) ".center(80, '='))
    coffs = list(isubtree_cutoffs(tree, root=1, **bounds))
    for i, (e, c) in enumerate(coffs):
        print(f"{i}.  {e[0]} -> {e[1]} : {c}")
    #
    print(f" Tabu cutoffs ({size}) ".center(80, '='))
    coffs = list(isubtree_sync_cutoffs(tree, root=1, size=size))
    for i, (cut, n_free, sync) in enumerate(coffs):
        print(f"{i}.  {cut[0]} -> {cut[1]} : {n_free}, {sync=}")
    if draw:
        draw_tree(tree, cuts=[c[0] for c in coffs])


def test_rand_cpu_cutoff(n: int = 10, workers: int = None, draw: bool = True):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_pseudo_mp"
    test_cpu_cutoff(tree, draw=False)
    print(f" CPU-based cutoffs ".center(80, '='))
    cuts = get_cpu_splits(tree, 1, workers)
    print(f"Calculated cuts({workers=}):", cuts)
    for i, c in enumerate(cuts):
        print(f"{i}.  {c[0]} -> {c[1]} ")
    if draw:
        draw_tree(tree, cuts=[c for c in cuts])


def test_rand_subtree_splits(n: int = 10, draw: bool = True):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_pseudo_mp"
    test_subtree_split(tree, draw=False)
    cuts = list(isubtree_splits(tree, root=1))
    print(f" Tabu cutoffs ".center(80, '='))
    print(f"Calculated cuts:", cuts)
    for i, (cut, sync) in enumerate(cuts):
        print(f"{i}.  {cut[0]} -> {cut[1]} : {sync=}")
    if draw:
        draw_tree(tree, cuts=[c[0] for c in cuts])


def run_test(tree: nx.DiGraph, M: int, L: int, root: int = 1, cp_end: int = None, delay: int = 10):
    partition, opt_cost, opt_lat = pseudo_mp_btree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    partition, opt_cost, opt_lat = pseudo_mp_ltree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    return partition, opt_cost, opt_lat


def test_ser_tree_pseudo_partitioning():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo_mp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=math.inf,
                  # L = 430,
                  delay=10)
    run_test(**params)


def test_random_tree_partitioning(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_pseudo_mp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  # L = 430,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_cpu_cutoff(cut_factor=3)
    # test_subtree_split(size=2)
    # test_subtree_split()
    # test_rand_cpu_cutoff(n=15, workers=3, draw=True)
    # test_rand_cpu_cutoff(n=15, workers=None, draw=True)
    # test_rand_subtree_splits(n=15, draw=True)

    test_ser_tree_pseudo_partitioning()
    # test_random_tree_partitioning()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import math
import sys
import warnings

import docplex
import docplex.cp.model as cpo
import docplex.mp.model as cpx
import networkx as nx

if sys.version_info.minor > 10 and docplex.docplex_version_minor <= 25:
    warnings.warn(f"docplex[{docplex.version.docplex_version_string}] package does not support Python version >3.10!")

from slambuc.alg import LP_LAT, INFEASIBLE, T_RESULTS, T_PART
from slambuc.alg.service import *
from slambuc.alg.tree.ser.ilp import ifeasible_subtrees
from slambuc.alg.util import (ibacktrack_chain, induced_subtrees, ser_subchain_latency, ser_subtree_cost,
                              recreate_subtree_blocks, verify_limits)
from slambuc.misc.util import get_cpo_path

CPO_PATH = get_cpo_path()


def build_tree_cfg_cpo_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                             L: int = math.inf, cpath: set[int] = frozenset(), delay: int = 1,
                             isubtrees: iter = ifeasible_subtrees) -> tuple[cpo.CpoModel, list[cpo.CpoIntVar]]:
    """
    Generate the configuration CP model.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = cpo.CpoModel(name="Tree_Partitioning")
    # Decision variables with precalculated coefficients
    c_x, l_x, X_n = [], [], collections.defaultdict(list)
    for i, (b, nodes) in enumerate(isubtrees(tree, root, M)):
        # Decision variable for the subtree
        x = cpo.binary_var(name=f"x_{b:02d}_{i}")
        # Add subtree block cost
        c_x.append(ser_subtree_cost(tree, b, nodes) * x)
        # Add subtree block latency if required
        if b in cpath:
            st_lat = ser_subchain_latency(tree, b, nodes, cpath)
            l_x.append((st_lat + delay) * x if b != root else st_lat * x)
        # Cache node coverage of subtree block
        for n in nodes:
            X_n[n].append(x)
    # Objective
    model.add(cpo.minimize(cpo.sum(c_x)))
    # Feasibility constraints
    for i, x_i in X_n.items():
        model.add(cpo.sum(x_i) == 1)
    # Latency constraint
    sum_lat = cpo.sum(l_x)
    model.add(sum_lat <= L if L < math.inf else sum_lat >= 0)
    model.add_kpi(sum_lat, LP_LAT)
    return model, X_n


def _set_cpo_context():
    """Set best performing CP solver configuration"""
    from docplex.cp.config import context
    context.verbose = 0
    context.model.add_source_location = False
    context.model.length_for_alias = 10
    context.model.name_all_constraints = False
    context.model.dump_directory = None
    context.model.sort_names = None
    context.solver.trace_cpo = False
    context.solver.trace_log = False
    context.solver.add_log_to_solution = False


def tree_cpo_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf, cp_end: int = None,
                          delay: int = 1, **kwargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a tree based on configuration CP formulation.

    :param tree:      service graph annotated with node runtime(ms), memory(MB) and edge rate
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:   invocation delay between blocks
    :return:        tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    model, Xn = build_tree_cfg_cpo_model(tree, root, M, L, cpath, delay)
    result = model.solve(agent='local', execfile=CPO_PATH, LogVerbosity='Quiet', SearchType='DepthFirst', Workers=1)
    if result:
        opt_cost, opt_lat = result.get_objective_values()[0], result.solution.get_kpis()[LP_LAT]
        return recreate_subtrees_from_cpo_xdict(tree, result, Xn), round(opt_cost, ndigits=6), round(opt_lat, ndigits=6)
    else:
        return INFEASIBLE


def recreate_subtrees_from_cpo_xdict(tree: nx.DiGraph, result: cpo.CpoSolveResult,
                                     Xn: dict[int, list[cpo.CpoIntVar]]) -> T_PART:
    """
    Extract barrier nodes from variable names (x_{b}_{w}) and recreate partitioning blocks.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param result:  result object
    :param Xn:      specific structure of decision variables
    :return:        calculated partitioning
    """
    barr = set(int(next(filter(lambda x: result[x], x_n)).name.split('_', 2)[1]) for x_n in Xn.values())
    return recreate_subtree_blocks(tree=tree, barr=barr)


########################################################################################################################


def build_greedy_tree_cplex_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                                  L: int = math.inf, cpath: set[int] = frozenset(),
                                  delay: int = 1) -> tuple[cpx.Model, dict[int, dict[int, docplex.mp.dvar]]]:
    """
    Generate the matrix ILP model using CPLEX Python binding.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = cpx.Model(name="Tree_Partitioning")
    # Empty decision variable matrix
    X = {j: {} for j in filter(lambda n: n is not PLATFORM, tree)}
    # Objective
    sum_cost = model.linear_expr()
    for j in X:
        cost_pre = 0
        nodes = set()
        for v in nx.dfs_preorder_nodes(tree, source=j):
            nodes |= {v}
            cost_vj = ser_subtree_cost(tree, j, nodes)
            X[v][j] = model.binary_var(f"x_{v:02d}_{j:02d}")
            sum_cost += (cost_vj - cost_pre) * X[v][j]
            cost_pre = cost_vj
    model.minimize(sum_cost)
    # Feasibility constraints
    for i in X:
        model.add_constraint(model.sum(X[i].values()) == 1, ctname=f"Cf_{i:02d}")
    # Knapsack constraints
    for j in X:
        model.add_constraint(model.sum(tree.nodes[i][MEMORY] * X[i][j]
                                       for i in nx.dfs_preorder_nodes(tree, source=j)) <= M, f"Ck_{j:02d}")
    # Connectivity constraints
    for j in X:
        for u, v in nx.dfs_edges(tree, source=j):
            model.add_constraint(X[u][j] - X[v][j] >= 0, ctname=f"Cc_{j:02d}_{u:02d}_{v:02d}")
    # Latency constraint
    sum_lat = model.linear_expr()
    for j in X:
        if j not in cpath:
            continue
        lat_pre = 0
        nodes = set()
        for v in nx.dfs_preorder_nodes(tree, source=j):
            nodes |= {v}
            # Add subtree block latency if required
            lat_vj = ser_subchain_latency(tree, j, nodes, cpath)
            sum_lat += (lat_vj - lat_pre) * X[v][j]
            if v == j and j != root:
                sum_lat += delay * X[v][j]
            lat_pre = lat_vj
    # Latency constraint
    if L < math.inf:
        model.add_constraint(sum_lat <= L, ctname=LP_LAT)
    # Add calculated sum latency as KPI for tracking
    model.add_kpi(sum_lat, LP_LAT)
    return model, X


def build_tree_cplex_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                           L: int = math.inf, cpath: set[int] = frozenset(),
                           delay: int = 1) -> tuple[cpx.Model, dict[int, dict[int, docplex.mp.dvar]]]:
    """
    Generate the matrix ILP model using CPLEX Python binding.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = cpx.Model(name="Tree_Partitioning")
    # Decision variable matrix with trivial variables
    X = {j: {j: model.binary_var(f"x_{j:02d}_{j:02d}")} for j in tree.nodes if j is not PLATFORM}
    # Empty objective
    sum_cost = model.linear_expr()
    # Empty latency constraint
    sum_lat = model.linear_expr()
    # Generate decision variables
    for (p, j), st_edges in induced_subtrees(tree, root):
        blk_mem = tree.nodes[j][MEMORY] * X[j][j]
        # Add coefficients for single node block [j]
        r_j, d_j, t_j = tree[p][j][RATE], tree[p][j][DATA], tree.nodes[j][RUNTIME]
        sum_cost += (r_j * (d_j + t_j) + sum(js[RATE] * js[DATA] for js in tree.succ[j].values())) * X[j][j]
        if j in cpath:
            jc = next(filter(lambda c: c in cpath, tree.successors(j)), None)
            sum_lat += (d_j + t_j + (math.ceil(tree[j][jc][RATE] / r_j) * tree[j][jc][DATA] if jc else 0)) * X[j][j]
            if j != root:
                sum_lat += delay * X[j][j]
        # Cache instance factor for nodes in cpath
        n_v = 1
        # Candidate nodes for block_j
        for u, v in st_edges:
            X[v][j] = model.binary_var(f"x_{v:02d}_{j:02d}")
            blk_mem += tree.nodes[v][MEMORY] * X[v][j]
            # Add coefficients for merging single node v to block [j,...]
            r_v, d_v, t_v = tree[u][v][RATE], tree[u][v][DATA], tree.nodes[v][RUNTIME]
            sum_cost += (r_v * (t_v - d_v) + sum(vs[RATE] * vs[DATA] for vs in tree.succ[v].values())) * X[v][j]
            # u -> v edge on cpath
            if v in cpath:
                n_v *= math.ceil(r_v / tree[next(tree.predecessors(u))][u][RATE])
                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                w_v = math.ceil(tree[v][vc][RATE] / r_v) * tree[v][vc][DATA] if vc else 0
                sum_lat += n_v * (t_v - d_v + w_v) * X[v][j]
            # Connectivity constraint
            model.add_constraint(X[u][j] - X[v][j] >= 0, ctname=f"Cc_{j:02d}_{u:02d}_{v:02d}")
        # Knapsack constraint, X[l][l] <= M for each leaf node l can be omitted
        if blk_mem.size > 1:
            model.add_constraint(blk_mem <= M, ctname=f"Ck_{j:02d}")
    # Objective
    model.minimize(sum_cost)
    # Feasibility constraints, X[root][root] = 1 can be omitted else it ensures that X[root][root] must be 1
    for i in X:
        model.add_constraint(model.sum(X[i].values()) == 1, ctname=f"Cf_{i:02d}")
    # Latency constraint
    if L < math.inf:
        model.add_constraint(sum_lat <= L, ctname=LP_LAT)
    # Add calculated sum latency as KPI for tracking
    model.add_kpi(sum_lat, LP_LAT)
    return model, X


def tree_cplex_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                            cp_end: int = None, delay: int = 1, **kwargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a tree based on matrix CPLEX ILP formulation.

    :param tree:      service graph annotated with node runtime(ms), memory(MB) and edge rate
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:   invocation delay between blocks
    :return:        tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    model, X = build_tree_cplex_model(tree, root, M, L, cpath, delay)
    solution = model.solve(agent='local', log_output=False, **kwargs)
    if solution is not None:
        opt_cost, opt_lat = solution.get_objective_value(), model.kpi_value_by_name(LP_LAT)
        return extract_subtrees_from_cplex_xmatrix(X), round(opt_cost, ndigits=6), round(opt_lat, ndigits=6)
    else:
        return INFEASIBLE


def extract_subtrees_from_cplex_xmatrix(X: dict[int, dict[int, docplex.mp.dvar]]) -> T_PART:
    """
    Extract barrier nodes from variable matrix(dict-of-dict) and recreate partitioning blocks.

    :param X:   specific structure of decision variables
    :return:    calculated partitioning
    """
    return [[i for i in sorted(X) if j in X[i] and round(X[i][j].solution_value, ndigits=6) == 1]
            for j in sorted(X) if round(X[j][j].solution_value, ndigits=6) == 1]



# Copyright 2024 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import heapq
import math

import networkx as nx

from slambuc.alg import T_RESULTS, INFEASIBLE
from slambuc.alg.util import block_cost, block_memory, block_cpu

# Naming convention for state-space DAG
START, END = 's', 't'


def encode_blk(b: int | str, w: int | str) -> str:
    """
    Encode node blocks as str.

    :param b:   barrier node
    :param w:   ending node
    :return:    encoded blk
    """
    return f"{b}|{w}"


def decode_blk(s: str) -> tuple[str, str]:
    """
    Decode encoded block str.

    :param s:   encoded str
    :return:    tuple of block barrier and ending nodes
    """
    return s.split('|', maxsplit=1)


def _build_sp_dag(runtime: list, memory: list, rate: list, M: int = math.inf, N: int = math.inf,
                  unit: int = 100) -> nx.DiGraph:
    """
    Build configuration state graph of the given function chain.

    :param runtime: Running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. Rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :return:        return state graph
    """
    n = len(runtime)
    # Initiate data structure for DAG
    _cache = collections.defaultdict(list)
    _cache.update({START: [START], END: [END]})
    dag = nx.DiGraph(directed=True)
    for b in range(n):
        prev = b - 1 if b > 0 else START
        for w in range(b, n):
            if block_memory(memory, b, w) > M or block_cpu(rate, b, w) > N:
                break
            blk_id = encode_blk(b, w)
            _cache[w].append(blk_id)
            blk_cost = block_cost(runtime, rate, b, w, unit)
            # Add connection between related subcases
            for p in _cache[prev]:
                dag.add_edge(p, blk_id, weight=blk_cost)
    # Add connection for ending subcases
    for p in _cache[n - 1]:
        dag.add_edge(p, END, weight=0)
    return dag


def hop_limited_shortest_path(dag: nx.DiGraph, source: str | int, target: str | int,
                              max_hops: int = math.inf) -> tuple[int, int, list[int | str]]:
    """
    Calculate the shortest path in graph 'dag' between 'source' and 'target' with the hop limit 'max_hops'.

    :param dag:         state graph
    :param source:      start node in 'dag'
    :param target:      end node in 'dag'
    :param max_hops:    hop limit of for the path
    :return:            calculated sum weights, hop count, nodes of the shortest path
    """
    #
    dist = collections.defaultdict(lambda: collections.defaultdict(lambda: math.inf))
    prev = collections.defaultdict(lambda: collections.defaultdict(lambda: None))
    pq = []
    dist[source][0] = 0
    heapq.heappush(pq, (source, 0))
    # Iterate over graph edges combined with possible hops
    while pq:
        u, h = heapq.heappop(pq)
        if h == max_hops:
            continue
        for v in dag.succ[u]:
            alt = dist[u][h] + dag[u][v]['weight']
            if alt < dist[v][h + 1]:
                dist[v][h + 1] = alt
                prev[v][h + 1] = u
                heapq.heappush(pq, (v, h + 1))
    # Reconstruct the shortest path
    if not len(dist[target]):
        return None, None, None
    path, _u, _h = [], target, min(dist[target], key=lambda x: dist[target][x])
    min_weight, hcount = dist[target][_h], _h
    while _u:
        path.append(_u)
        _u = prev[_u][_h]
        _h -= 1
    path.reverse()
    return min_weight, hcount, path


def sp_chain_partitioning(runtime: list, memory: list, rate: list, M: int = math.inf, N: int = math.inf,
                          L: int = math.inf, delay: int = 1, unit: int = 100, **kwargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a chain based on the node properties of *running time*, *memory usage* and
    *invocation rate* with respect to an upper bound **M** on the total memory of blocks and a latency constraint **L**
    defined on the subchain between *start* and *end* nodes.

    Partitioning is based on the shortest path calculation of the state graph of feasible blocks.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :return:        tuple of barrier nodes, sum cost of the partitioning, and the calculated edge cuts
    """
    dag = _build_sp_dag(runtime, memory, rate, M, N, unit)
    h_max = math.floor(min((L - sum(runtime)) / delay, len(runtime) - 1))
    min_cost, hops, sp = hop_limited_shortest_path(dag, source=START, target=END, max_hops=h_max + 2)
    return ([int(decode_blk(v)[0]) for v in sp[1:-1]], min_cost, hops - 2) if sp else INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import sys
import time

import networkx as nx
import pytest

from slambuc.alg import LP_LAT
from slambuc.alg.service import NAME
from slambuc.alg.tree.ser.ilp_cplex import (build_greedy_tree_cplex_model, build_tree_cplex_model,
                                            tree_cplex_partitioning, extract_subtrees_from_cplex_xmatrix,
                                            build_tree_cfg_cpo_model, CPO_PATH, recreate_subtrees_from_cpo_xdict)
from slambuc.alg.util import ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import (evaluate_ser_tree_partitioning, print_var_matrix, convert_var_dict,
                               print_cplex_matrix_values, get_cpo_path, get_cplex_path)


@pytest.mark.skipif(get_cpo_path() is None, reason="CPO is not available!")
def test_cpo_model_creation(save_file: bool = False):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-cplex_ser"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model, X = build_tree_cfg_cpo_model(**params)
    _d = time.perf_counter() - _s
    print(f"Greedy CPO Model building time: {_d * 1000} ms")
    #
    print("Summary:")
    model.print_information()
    print("Model:")
    model.export_model(short_output=True)
    if save_file:
        model.export_model("chain_model.lp")


@pytest.mark.skipif(get_cpo_path() is None, reason="CPO is not available!")
def test_cpo_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-cplex_ser"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  # L=math.inf,
                  L=430,
                  delay=10)
    model, Xn = build_tree_cfg_cpo_model(**params)
    print("CPO Greedy model")
    model.print_information()
    print("Solve model...")
    solution = model.solve(agent='local', execfile=CPO_PATH,
                           TimeMode='ElapsedTime', LogVerbosity='Verbose', WarningLevel=3,
                           SearchType='DepthFirst', Workers=1)
    print("Solve status:", solution.get_solve_status())
    print("Solution time:", solution.get_solve_time())
    print("Result:")
    solution.print_solution()
    print(f"Cost/Lat: {solution.get_objective_values()[0]} / {solution.get_kpis()[LP_LAT]}")
    _s = time.perf_counter()
    partition = recreate_subtrees_from_cpo_xdict(tree, solution, Xn)
    _d = time.perf_counter() - _s
    print(f"Recreate:  {_d * 1000} ms")
    print(f"Partitioning: {partition = }")


########################################################################################################################


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
def test_cplex_model_creation(save_file: bool = False):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-cplex_ser"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model_greedy, X_greedy = build_greedy_tree_cplex_model(**params)
    _d = time.perf_counter() - _s
    print(f"Greedy Model building time: {_d * 1000} ms")
    _s = time.perf_counter()
    model, X = build_tree_cplex_model(**params)
    _d = time.perf_counter() - _s
    print(f"Direct Model building time: {_d * 1000} ms")
    X = convert_var_dict(X)
    X_greedy = convert_var_dict(X_greedy)
    print("  Decision variables  ".center(80, '='))
    print("Greedy:")
    print_var_matrix(X_greedy)
    print("Direct:")
    print_var_matrix(X)
    #
    print("Summary:")
    model.print_information()
    if save_file:
        model.export_as_lp("tree_cplex_model.lp")


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_cplex_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-cplex_ser"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  # L=math.inf,
                  L=430,
                  delay=10)
    print("Test greedy model")
    model, X = build_greedy_tree_cplex_model(**params)
    model.solve(log_output=True)
    model.print_solution()
    print("  build CPLEX model  ".center(80, '='))
    model, X = build_tree_cplex_model(**params)
    model.print_information()
    print("  Run CPLEX solver  ".center(80, '='))
    solution = model.solve(log_output=True)
    print(f"Cost/Lat: {solution.get_objective_value()} / {model.kpi_value_by_name(LP_LAT)}")
    print("Model report:")
    model.report()
    #
    print("Model solution:")
    model.print_solution()
    print("Solution details:")
    print(model.get_solve_details())
    print_cplex_matrix_values(convert_var_dict(X))
    _s = time.perf_counter()
    ext_partition = extract_subtrees_from_cplex_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {ext_partition = }")


########################################################################################################################


def run_test(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, delay: int):
    partition, opt_cost, opt_lat = tree_cplex_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    return partition, opt_cost, opt_lat


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_ser_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-cplex_ser"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  # L=math.inf,
                  L=430,
                  delay=10)
    run_test(**params)


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_random_ser_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-cplex_ser"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_cpo_model_creation()
    test_cpo_model_solution()
    #
    # test_cplex_model_creation(save_file=False)
    # test_cplex_model_solution()
    #
    # test_ser_tree()
    # test_random_ser_tree()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

from slambuc.alg import INFEASIBLE, T_RESULTS, T_PART_GEN
from slambuc.alg.util import ipowerset, split_chain, block_memory, block_cost, block_latency, block_cpu


def ichain_blocks(memory: list[int], rate: list[int], N: int, M: int) -> T_PART_GEN:
    """
    Calculates all combinations of chain cuts with respect to *memory* and *rate* values and the constraint **M**.

    The calculation is improved compared to brute force to only start calculating cuts from minimal cut size *c_min*.

    :param memory:  list of node memory values
    :param rate:    list of invocation rate values
    :param N:       number of vCPU cores
    :param M:       upper memory limit
    :return:        Generator over M-feasible cuts.
    """
    n = len(memory)
    for cut in ipowerset(range(1, n), start=math.ceil(sum(memory) / M) - 1):
        barr = sorted({0}.union(cut))
        # Consider only block with the appropriate size
        valid = [blk for blk in split_chain(barr, n)
                 if block_memory(memory, blk[0], blk[-1]) <= M and block_cpu(rate, blk[0], blk[-1]) <= N]
        if len(valid) == len(barr):
            yield valid


def greedy_chain_partitioning(runtime: list[int], memory: list[int], rate: list[int], M: int = math.inf,
                              N: int = math.inf, L: int = math.inf, start: int = 0, end: int = None,
                              delay: int = 1, unit: int = 100) -> list[T_RESULTS]:
    """
    Calculates all minimal-cost partitioning outcomes of a given chain by applying exhaustive search.

    Parameters are the same as the partitioning algorithms in ``slambuc.alg.chain.dp.mtx``
    and ``slambuc.alg.chain.dp.min``.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param start:   head node of the latency-limited subchain
    :param end:     tail node of the latency-limited subchain
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :return:        list if min-cost partitions, related optimal cost and latency
    """
    best_res, best_cost = [INFEASIBLE], math.inf
    for partition in ichain_blocks(memory, rate, N, M):
        if (sum_lat := sum(block_latency(runtime, blk[0], blk[-1], delay, start, end) for blk in partition)) > L:
            continue
        elif (sum_cost := sum(block_cost(runtime, rate, blk[0], blk[-1], unit) for blk in partition)) == best_cost:
            best_res.append((partition, sum_cost, sum_lat))
        elif sum_cost < best_cost:
            best_res, best_cost = [(partition, sum_cost, sum_lat)], sum_cost
    return best_res



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import time

import networkx as nx
import pandas as pd

from slambuc.alg.service import NAME
from slambuc.alg.tree.ser.bicriteria import biheuristic_btree_partitioning, bifptas_ltree_partitioning, \
    bifptas_dual_ltree_partitioning
from slambuc.alg.tree.ser.pseudo import pseudo_btree_partitioning, pseudo_ltree_partitioning
from slambuc.alg.util import recalculate_partitioning
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import evaluate_ser_tree_partitioning


def run_validation_test(tree: nx.DiGraph, M: int, L: int, root: int = 1, cp_end: int = None, delay: int = 10):
    print(" Run pseudo BTree algorithm ".center(80, '#'))
    b_part, b_opt_cost, b_opt_lat = pseudo_btree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, b_part, b_opt_cost, b_opt_lat, root, cp_end, M, L, delay)
    print(" Run bi-PTAS BTree algorithm ".center(80, '#'))
    bc_part, bc_opt_cost, bc_opt_lat = biheuristic_btree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, bc_part, bc_opt_cost, bc_opt_lat, root, cp_end, M, L, delay)
    print(" Run pseudo LTree algorithm ".center(80, '#'))
    l_part, l_opt_cost, l_opt_lat = pseudo_ltree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, l_part, l_opt_cost, l_opt_lat, root, cp_end, M, L, delay)
    print(" Run bi-PTAS LTree algorithm ".center(80, '#'))
    lc_part, lc_opt_cost, lc_opt_lat = bifptas_ltree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, lc_part, lc_opt_cost, lc_opt_lat, root, cp_end, M, L, delay)


def test_ser_bicriteria_tree_partitioning():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_bic"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  # L=math.inf,
                  L=430,
                  delay=10)
    run_validation_test(**params)


def test_random_bicriteria_tree_partitioning(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_bic"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  # L = 430,
                  delay=10)
    run_validation_test(**params)


def get_accuracy_stats(n: int = 10, M: int = 6, L: int = math.inf, Epsilon: float = 0.0, Lambda: float = 0.0,
                       stop_failed: bool = False):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_bic"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=M,
                  # L=math.inf,
                  L=L,
                  delay=10,
                  bidirectional=False)
    bic_params = params.copy()
    bic_params.update(Epsilon=Epsilon, Lambda=Lambda)
    #
    # Btree exact
    _s = time.perf_counter()
    btree_part, btree_opt_cost, btree_opt_lat = pseudo_btree_partitioning(**params)
    btree_time = time.perf_counter() - _s
    _s = time.perf_counter()
    # Bi-Btree reference
    b_part, b_opt_cost, b_opt_lat = biheuristic_btree_partitioning(**params)
    bc_exact_time = time.perf_counter() - _s
    _s = time.perf_counter()
    # Bi-Btree
    bc_part, bc_opt_cost, bc_opt_lat = biheuristic_btree_partitioning(**bic_params)
    bc_bic_time = time.perf_counter() - _s
    #
    # Ltree exact
    _s = time.perf_counter()
    ltree_part, ltree_opt_cost, ltree_opt_lat = pseudo_ltree_partitioning(**params)
    ltree_time = time.perf_counter() - _s
    # Bi-Ltree reference
    _s = time.perf_counter()
    l_part, l_opt_cost, l_opt_lat = bifptas_ltree_partitioning(**params)
    lc_exact_time = time.perf_counter() - _s
    _s = time.perf_counter()
    # Bi-Ltree
    lc_part, lc_opt_cost, lc_opt_lat = bifptas_ltree_partitioning(**bic_params)
    lc_bic_time = time.perf_counter() - _s
    #
    # Bi-Ltree dual reference
    _s = time.perf_counter()
    dl_part, dl_opt_cost, dl_opt_lat = bifptas_dual_ltree_partitioning(**params)
    dlc_exact_time = time.perf_counter() - _s
    _s = time.perf_counter()
    # Bi-Ltree dual
    dlc_part, dlc_opt_cost, dlc_opt_lat = bifptas_dual_ltree_partitioning(**bic_params)
    dlc_bic_time = time.perf_counter() - _s
    #
    print(" Results ".center(80, '#'))
    print(f"Btree:     {b_part}")
    print(f"Bi-Btree:  {bc_part}")
    print(f"Ltree:     {l_part}")
    print(f"Bi-Ltree:  {lc_part}")
    print(f"DLtree:    {l_part}")
    print(f"Bi-DLtree: {lc_part}")
    if btree_part:
        print("#### Btree alg execution:")
        print(f"Btree pseudo exact cost:  {btree_opt_cost:>5},  lat: {btree_opt_lat:>5}  "
              f"with time: {btree_time * 1000:>3.4f} ms")
        print(f"Bi-Btree ref. exact cost: {b_opt_cost:>5},  lat: {b_opt_lat:>5}  "
              f"with time: {bc_exact_time * 1000:>3.4f} ms")
    if bc_part:
        print(f"Bi-Btree bicriteria cost: {bc_opt_cost:>5},  lat: {bc_opt_lat:>5}  "
              f"with time: {bc_bic_time * 1000:>3.4f} ms")
        bc_sum_cost, bc_sum_lat = recalculate_partitioning(tree, bc_part, root=params['root'], cp_end=params['cp_end'],
                                                           delay=params['delay'])
        b_cost_err = (bc_sum_cost - btree_opt_cost) / btree_opt_cost
        b_lat_err = (bc_sum_lat - L) / L
        print(f">>>> Errors:   {b_cost_err = :.4f}  (<= {Epsilon=}) [{bc_sum_cost}/{btree_opt_cost=}], "
              f"   {b_lat_err = :.4f}  (<= {Lambda=}) [{bc_sum_lat}/{L=}]")
    else:
        b_cost_err, b_lat_err = 0, 0
    if ltree_part:
        print("#### Ltree alg execution:")
        print(f"Ltree pseudo exact cost:  {ltree_opt_cost:>5},  lat: {ltree_opt_lat:>5},  "
              f"with time: {ltree_time * 1000:>3.4f} ms")
    if lc_part:
        print(f"Bi-Ltree ref. exact cost: {l_opt_cost:>5},  lat: {lc_opt_lat:>5},  "
              f"with time: {lc_exact_time * 1000:>3.4f} ms")
        print(f"Bi-Ltree bicriteria cost: {lc_opt_cost:>5},  lat: {lc_opt_lat:>5},  "
              f"with time: {lc_bic_time * 1000:>3.4f} ms")
        lc_sum_cost, lc_sum_lat = recalculate_partitioning(tree, lc_part, root=params['root'], cp_end=params['cp_end'],
                                                           delay=params['delay'])
        l_cost_err = (lc_sum_cost - ltree_opt_cost) / ltree_opt_cost
        l_lat_err = (lc_sum_lat - L) / L
        print(f">>>> Errors:    {l_cost_err = :.4f}  (<= {Epsilon=}) [{lc_sum_cost}/{ltree_opt_cost=}], "
              f"   {l_lat_err = :.4f}  (<= {Lambda=}) [{lc_sum_lat}/{L=}]")
    else:
        l_cost_err, l_lat_err = 0, 0
    if dlc_part:
        print(f"Bi-Ltree dual ref. cost:  {dl_opt_cost:>5},  lat: {dlc_opt_lat:>5},  "
              f"with time: {dlc_exact_time * 1000:>3.4f} ms")
        print(f"Bi-Ltree dual bic. cost:  {dlc_opt_cost:>5},  lat: {dlc_opt_lat:>5},  "
              f"with time: {dlc_bic_time * 1000:>3.4f} ms")
        dlc_sum_cost, dlc_sum_lat = recalculate_partitioning(tree, dlc_part, root=params['root'],
                                                             cp_end=params['cp_end'], delay=params['delay'])
        dl_cost_err = (dlc_sum_cost - ltree_opt_cost) / ltree_opt_cost
        dl_lat_err = (dlc_sum_lat - L) / L
        print(f">>>> Errors:   {dl_cost_err = :.4f}  (<= {Epsilon=}) [{dlc_sum_cost}/{ltree_opt_cost=}], "
              f"  {dl_lat_err = :.4f}  (<= {Lambda=}) [{dlc_sum_lat}/{L=}]")
    else:
        dl_cost_err, dl_lat_err = 0, 0
    if stop_failed:
        assert all((b_cost_err <= Epsilon if math.isfinite(b_cost_err) else True,
                    l_cost_err <= Epsilon if math.isfinite(l_cost_err) else True,
                    dl_cost_err <= Epsilon if math.isfinite(dl_cost_err) else True,
                    b_lat_err <= Lambda if math.isfinite(b_lat_err) else True,
                    dl_lat_err <= Lambda if math.isfinite(dl_lat_err) else True,
                    l_lat_err <= Lambda if math.isfinite(l_lat_err) else True))
    return ((btree_time * 1000, bc_exact_time * 1000, bc_bic_time * 1000, b_cost_err, b_lat_err),  # Btree stats
            (ltree_time * 1000, lc_exact_time * 1000, lc_bic_time * 1000, l_cost_err, l_lat_err),
            (ltree_time * 1000, dlc_exact_time * 1000, dlc_bic_time * 1000, dl_cost_err, dl_lat_err))  # Ltree stats


def test_accuracy():
    print(get_accuracy_stats())


def stress_test(iteration: int = 100, n: int = 10, M: int = 6, L: int = math.inf, Epsilon: float = 0.5,
                Lambda: float = 0.5, stop_failed: bool = False):
    b_stat, l_stat, dl_stat = zip(*[get_accuracy_stats(n, M, L, Epsilon, Lambda, stop_failed=stop_failed)
                                    for _ in range(iteration)])
    b_df = pd.DataFrame(b_stat, columns=['Btree', 'BiBtree_exact_time', 'BiBtree_time', 'B_omega', 'B_lambda'])
    l_df = pd.DataFrame(l_stat, columns=['Ltree', 'BiLtree_exact_time', 'BiLtree_time', 'L_omega', 'L_lambda'])
    dl_df = pd.DataFrame(dl_stat, columns=['Ltree', 'BiDLtree_exact_time', 'BiDLtree_time',
                                           'Dual_L_omega', 'Dual_L_lambda'])
    pd.set_option('display.expand_frame_repr', False)
    print(" Btree stat ".center(80, '#'))
    print(b_df.describe().transpose())
    print(" Ltree stat ".center(80, '#'))
    print(l_df.describe().transpose())
    print(" Dual Ltree stat ".center(80, '#'))
    print(dl_df.describe().transpose())


if __name__ == '__main__':
    # test_ser_bicriteria_tree_partitioning()
    # test_random_bicriteria_tree_partitioning()
    # test_accuracy(n=10, M=6, L=450, Epsilon=0.5, Lambda=0.2, stop_failed=True)
    # test_accuracy(n=10, M=6, L=math.inf, Epsilon=0.5, Lambda=0.2, stop_failed=True)
    stress_test(iteration=100, n=20, M=20, L=855, Epsilon=0, Lambda=0, stop_failed=True)
    # stress_test(iteration=100, n=40, M=30, L=855, Epsilon=0.5, Lambda=0.2, stop_failed=True)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import importlib.resources
import itertools
import math
import pickle
import random

import pandas as pd
import scipy

# Random job generation source code is moved from package *spar* with adaptations to newer versions of Python3.11
# and Scipy 1.10.  See also: https://github.com/All-less/trace-generator/blob/master/spar/generate.py

HIST_DATA_DIR = importlib.resources.files("slambuc.gen.cluster").joinpath("hist")
DIST_CACHE = {}


def draw(hist_name: str, num: int = 1, path: list = tuple(), ndigits: int = 2, positive: bool = True,
         output_integer: bool = False, seed: int = None) -> list[int | float]:
    """
    Draw random samples from a given distribution.

    Random job generation source code is moved from package *spar* with adaptations to newer versions of Python3.11
    and Scipy 1.10.  See also: https://github.com/All-less/trace-generator/blob/master/spar/generate.py
    """
    if hist_name not in DIST_CACHE:
        with (HIST_DATA_DIR / f"{hist_name}.pkl").open('rb') as f:
            hist = pickle.load(f)
            if isinstance(hist, dict):
                DIST_CACHE[hist_name] = {num: scipy.stats.rv_histogram(h, density=False, seed=seed)
                                         for num, h in hist.items()}
            else:
                DIST_CACHE[hist_name] = scipy.stats.rv_histogram(hist, density=False, seed=seed)
    dist = DIST_CACHE[hist_name]
    for p in path:
        dist = dist[p]
    samples = []
    while len(samples) < num:
        data = round(dist.rvs(), ndigits=None if output_integer else ndigits)
        if positive and data <= 0:
            continue
        samples.append(data)
    return samples


def random_levels(num_nodes: int, seed: int) -> list[int]:
    cpl = min(num_nodes, draw("cpl_hist", path=(min(num_nodes, 35),), output_integer=True, seed=seed)[0])
    levels = draw("level_hist", num=num_nodes - cpl, output_integer=True, path=(min(cpl, 20),), seed=seed)
    levels.extend(range(1, cpl + 1))
    return levels


def random_dag(num_nodes: int, seed: int = None) -> dict[int, list[int]]:
    if num_nodes == 1:
        return {1: []}
    # randomly select a critical path length and assign nodes along it
    nodes = collections.defaultdict(list)
    for n, l in enumerate(sorted(random_levels(num_nodes, seed=seed)), start=1):
        nodes[l].append(n)
    # randomly generate edges
    parents = {n: [] for n in range(1, num_nodes + 1)}
    for l in range(1, len(nodes)):
        for n in nodes[l]:
            for c in set(random.sample(nodes[l + 1], math.ceil(len(nodes[l + 1]) / len(nodes[l]) * 3 / 4))):
                parents[c].append(n)
    return parents


def random_job(task_num: int = None, seed: int = None) -> pd.DataFrame:
    random.seed(seed)
    if task_num is None:
        task_num = draw("task_num_hist", num=1, seed=seed, output_integer=True)[0]
    job_dag = random_dag(task_num, seed)  # { <task_1>: [ <parent_1, parent_2>, ... ], ... }
    # generate task_name, duration, plan_cpu, plan_mem, inst_num for each task
    task_info = zip(["_".join(itertools.chain((f"T{k}",), map(str, v))) for k, v in job_dag.items()],
                    draw("task_duration_hist", num=task_num, seed=seed, output_integer=True),
                    draw("task_cpu_hist", num=task_num, seed=seed),
                    draw("task_mem_hist", num=task_num, seed=seed),
                    draw("instance_num_hist", num=task_num, seed=seed, output_integer=True))
    return pd.DataFrame(task_info, columns=('task', 'duration', 'cpu', 'mem', 'num'))



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
from collections.abc import Generator

import networkx as nx
import pulp as lp

from slambuc.alg import LP_LAT, INFEASIBLE, T_FRESULTS, T_FPART
from slambuc.alg.service.common import *
from slambuc.alg.util import (ipowerset, ipostorder_dfs, ibacktrack_chain, gen_subtree_memory, gen_subtree_cost,
                              gen_subchain_latency, recreate_subtree_blocks, x_eval)


def ifeasible_gen_subtrees(tree: nx.DiGraph, root: int, M: int, N: int = 1) -> Generator[tuple[int, set[int]]]:
    """
    Generate M-feasible(connected) subtrees and roots in bottom-up way, which meet the memory constraint *M*.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :return:        generator of subtree root and regarding subtree nodes
    """
    subtrees = collections.defaultdict(list)
    for _, n in ipostorder_dfs(tree, root):
        for children in ipowerset(tuple(tree.successors(n))):
            for sts in itertools.product(*(subtrees[c] for c in children)):
                st = {n}.union(*sts)
                if gen_subtree_memory(tree, n, st, N) <= M:
                    subtrees[n].append(st)
                    yield n, st
        for c in tree.succ[n]:
            del subtrees[c]


def build_gen_tree_cfg_model(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                             exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t,
                             L: int = math.inf, cp_end: int = None,
                             delay: int = 1) -> tuple[lp.LpProblem, list[lp.LpVariable]]:
    """
    Generate the configuration ILP model with the given *flavors*.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Tree_Partitioning", sense=lp.LpMinimize)
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Decision variables with precalculated coefficients
    c_x, l_x, X = [], [], {f: {v: list() for v in tree if v is not PLATFORM} for f in flavors}
    for fi, f in enumerate(flavors):
        for i, (b, nodes) in enumerate(ifeasible_gen_subtrees(tree, root, f.mem, f.ncore)):
            # Decision variable for the subtree
            x = lp.LpVariable(f"x_{fi}_{b:02d}_{i}", cat=lp.LpBinary)
            # Add subtree block cost
            c_x.append(gen_subtree_cost(tree, b, nodes, f.ncore, exec_calc) * x)
            # Add subtree block latency if required
            if b in cpath:
                st_lat = gen_subchain_latency(tree, b, nodes, cpath, f.ncore, exec_calc)
                l_x.append((st_lat + delay) * x if b != root else st_lat * x)
            # Cache node coverage of subtree block
            for n in nodes:
                X[f][n].append(x)
    # Objective
    model += lp.lpSum(c_x)
    # Feasibility constraints
    for v in tree:
        if v is PLATFORM:
            continue
        model += lp.lpSum(X[f][v] for f in X) == 1, f"Cf_{v:03d}"
    # Latency constraint
    model += lp.lpSum(l_x) <= L if L < math.inf else lp.lpSum(l_x) >= 0, LP_LAT
    return model, X


def tree_gen_hybrid_partitioning(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                                 exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t,
                                 L: int = math.inf, cp_end: int = None, delay: int = 1, solver: lp.LpSolver = None,
                                 timeout: int = None, **lpargs) -> T_FRESULTS:
    """
    Calculate minimal-cost partitioning of a tree based on configuration LP formulation and given *flavors*.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param flavors:     list of flavors resources given by the tuple of available *(memory, relative CPU cores)*
    :param exec_calc:   function that calculates the effective runtimes from reference runtime and available CPU cores
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:       invocation delay between blocks
    :param solver:      specific solver class (default: COIN-OR CBC)
    :param timeout:     time limit in sec
    :param lpargs:      additional LP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    model, X = build_gen_tree_cfg_model(tree, root, flavors, exec_calc, L, cp_end, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = lp.value(model.objective), lp.value(model.constraints[LP_LAT])
        return recreate_st_from_gen_xdict(tree, X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def recreate_st_from_gen_xdict(tree: nx.DiGraph, X: dict[Flavor, dict[int, list[lp.LpVariable]]]) -> T_FPART:
    """
    Extract barrier nodes from variable names (x_{b}_{w}) and recreate partitioning blocks.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param X:       internal structure of decision variables
    :return:        partition blocks
    """
    dbarr = {int(x.name.rsplit('_', 2)[1]): f
             for f in X for v in X[f] for x in filter(lambda _x: _x.varValue == 1, X[f][v])}
    return [(p, dbarr[p[0]]) for p in recreate_subtree_blocks(tree=tree, barr=dbarr)]


########################################################################################################################


def build_gen_tree_mtx_model(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                             exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t,
                             L: int = math.inf, cp_end: int = None, subchains: bool = False,
                             delay: int = 1) -> tuple[lp.LpProblem, dict[int, dict[int, dict[int, lp.LpVariable]]]]:
    """
    Generate the matrix ILP model with the given *flavors*.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Tree_Partitioning", sense=lp.LpMinimize)
    # Empty decision variable matrix
    X = {f: {j: {} for j in tree if j is not PLATFORM} for f in flavors}
    # Objective
    sum_cost = lp.LpAffineExpression()
    for fi, f in enumerate(flavors):
        for j in X[f]:
            cost_pre = 0
            nodes = set()
            for v in nx.dfs_preorder_nodes(tree, source=j):
                nodes |= {v}
                cost_vj = gen_subtree_cost(tree, j, nodes, f.ncore, exec_calc)
                X[f][v][j] = lp.LpVariable(f"x_{fi}_{v:02d}_{j:02d}", cat=lp.LpBinary)
                sum_cost += (cost_vj - cost_pre) * X[f][v][j]
                cost_pre = cost_vj
    model += sum_cost
    # Feasibility constraints
    for v in tree:
        if v is PLATFORM:
            continue
        model += lp.lpSum(X[f][v] for f in flavors) == 1, f"Cf_{v:02d}"
    # Knapsack constraints
    for fi, f in enumerate(X):
        for j in X[f]:
            # Cumulative memory demand of prefetched models
            model += lp.lpSum(tree.nodes[i][MEMORY] * X[f][i][j]
                              for i in nx.dfs_preorder_nodes(tree, source=j)) <= f.mem, f"C_{fi}_k{j:02d}"
            r_j = tree[next(tree.predecessors(j))][j][RATE]
            # Operative memory demand of instances running in parallel
            for u, v in nx.dfs_edges(tree, source=j):
                vj_sat = min(math.ceil(tree[u][v][RATE] / r_j), math.ceil(f.ncore / tree.nodes[v].get(CPU, 1)))
                # Add only non-trivial memory constraint
                if vj_sat > 1:
                    model += vj_sat * tree.nodes[v][MEMORY] * X[f][v][j] <= f.mem, f"Ck_{fi}_{j:02d}_{v:02d}"
    # Connectivity constraints
    for fi, f in enumerate(X):
        for j in X[f]:
            for u, v in nx.dfs_edges(tree, source=j):
                model += X[f][u][j] - X[f][v][j] >= 0, f"Cc_{fi}_{j:02d}_{u:02d}_{v:02d}"
    # Path-tree constraints
    if subchains:
        for fi, f in enumerate(X):
            for pt in ((lp.lpSum(X[f][i][j] for i in tree.successors(v)) <= 1, f"Cp_{fi}_{j:02d}_{v:02d}")
                       for v in X[f] for j in X[f][v] if tree.succ[v]):
                model += pt
    # Latency constraint
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    sum_lat = lp.LpAffineExpression()
    for fi, f in enumerate(X):
        for j in X[f]:
            if j not in cpath:
                continue
            lat_pre = 0
            nodes = set()
            for v in nx.dfs_preorder_nodes(tree, source=j):
                nodes |= {v}
                # Add subtree block latency if required
                lat_vj = gen_subchain_latency(tree, j, nodes, cpath, f.ncore, exec_calc)
                sum_lat += (lat_vj - lat_pre) * X[f][v][j]
                if v == j and j != root:
                    sum_lat += delay * X[f][v][j]
                lat_pre = lat_vj
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def tree_gen_mtx_partitioning(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                              exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t,
                              L: int = math.inf, cp_end: int = None, subchains: bool = False, delay: int = 1,
                              solver: lp.LpSolver = None, timeout: int = None, **lpargs) -> T_FRESULTS:
    """
    Calculate minimal-cost partitioning of a tree based on matrix LP formulation and given *flavors*.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param flavors:     list of flavors resources given by the tuple of available *(memory, relative CPU cores)*
    :param exec_calc:   function that calculates the effective runtimes from reference runtime and available CPU cores
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param subchains:   only subchain blocks are considered (path-tree)
    :param delay:       invocation delay between blocks
    :param solver:      specific solver class (default: COIN-OR CBC)
    :param timeout:     time limit in sec
    :param lpargs:      additional LP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    model, X = build_gen_tree_mtx_model(tree, root, flavors, exec_calc, L, cp_end, subchains, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = round(lp.value(model.objective), 0), round(lp.value(model.constraints[LP_LAT]), 0)
        return extract_st_from_gen_xmatrix(X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def extract_st_from_gen_xmatrix(X: dict[Flavor, dict[int, dict[int, lp.LpVariable]]]) -> T_FPART:
    """
    Extract barrier nodes from variable matrix(dict-of-dict) and recreate partitioning blocks.

    :param X:       internal structure of decision variables
    :return:        partition blocks
    """
    return sorted(([i for i in sorted(X[f]) if j in X[f][i] and x_eval(X[f][i][j])], f)
                  for f in X for j in sorted(X[f]) if x_eval(X[f][j][j]))


########################################################################################################################


def all_gen_tree_mtx_partitioning(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                                  exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t,
                                  L: int = math.inf, cp_end: int = None, subchains: bool = False, delay: int = 1,
                                  solver: lp.LpSolver = None, timeout: int = None, **lpargs) -> list[T_FPART]:
    """
    Calculate all minimal-cost partitioning variations of a tree based on matrix ILP formulation and *flavors*.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param flavors:     list of flavors resources given by the tuple of available *(memory, relative CPU cores)*
    :param exec_calc:   function that calculates the effective runtimes from reference runtime and available CPU cores
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param subchains:   only subchain blocks are considered (path-tree)
    :param delay:       invocation delay between blocks
    :param solver:      specific solver class (default: COIN-OR CBC)
    :param timeout:     time limit in sec
    :param lpargs:      additional LP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Get model
    model, X = build_gen_tree_mtx_model(tree, root, flavors, exec_calc, L, cp_end, subchains, delay)
    # Init for min-cost results
    results, min_cost = [], math.inf
    solver = solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs)
    # Iterate over the optimal solutions
    while model.solve(solver) == lp.LpStatusOptimal:
        opt_cost, opt_lat = round(model.objective.value(), 0), round(model.constraints[LP_LAT].value(), 0)
        # If solution is not min-cost then exit else store the min-cost solution
        if min_cost < opt_cost:
            break
        else:
            min_cost = opt_cost
        results.append((extract_st_from_gen_xmatrix(X), opt_cost, L + opt_lat if L < math.inf else opt_lat))
        # Collect barrier nodes of the current optimal solution
        barr = [(f, j) for f in X for j in X[f] if x_eval(X[f][j][j])]
        # Add extra constraint for excluding the current set of barrier nodes
        model += (lp.lpSum(X[f][j][j] for f, j in barr) <= len(barr) - 1,
                  f"Ca_{'_'.join(map(str, itertools.chain.from_iterable(barr)))}")
    # Return min-cost solutions
    return results if results else [INFEASIBLE]



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import pathlib
import random

from slambuc.gen.io import save_trees_to_file
from slambuc.misc.generator import get_random_tree

# Attribute intervals
RUNTIME = (10, 500)  # Overall running time in ms
MEMORY = (64, 512)  # Peak memory demand in MB
DATA = (10, 100)  # Read/write overhead in ms
RATE = (1, 10)  # Invocations rate in 1/s

DEF_RAND_TREE_PREFIX = "random_tree"


def generate_random_trees(n: int, data_dir: str, iteration: int = 1000, file_prefix: str = DEF_RAND_TREE_PREFIX):
    """
    Generate random trees with attributes uniformly drawn form intervals.

    :param n:           tree size
    :param data_dir:    directory of saved trees
    :param iteration:   number of generated trees
    :param file_prefix: prefix name of tree files
    """
    print(f"Generating random trees with size {n}...")
    trees = [get_random_tree(nodes=n, runtime=RUNTIME, memory=MEMORY, rate=RATE, data=DATA) for _ in range(iteration)]
    file_name = pathlib.Path(data_dir, f"{file_prefix}_n{n}.npy").resolve()
    print(f"Saving trees into {file_name}...")
    save_trees_to_file(trees, file_name)


def generate_all_random_trees(data_dir: str, iteration: int = 100, start: int = 10, end: int = 100, step: int = 10,
                              file_prefix: str = DEF_RAND_TREE_PREFIX):
    """
    Generate random service trees with random sizes from given intervals.

    :param data_dir:    directory of saved trees
    :param iteration:   number of generated trees
    :param start:       minimum of size intervals
    :param end:         maximum of size intervals
    :param step:        step size of intervals
    :param file_prefix: prefix name of tree files
    """
    for min_size, max_size in itertools.pairwise(range(start, end + step, step)):
        print(f"Generating random trees with {min_size} <= size <= {max_size}...")
        trees = [get_random_tree(nodes=random.randint(min_size, max_size), runtime=RUNTIME, memory=MEMORY,
                                 rate=RATE, data=DATA, name=DEF_RAND_TREE_PREFIX + f"_{i}") for i in range(iteration)]
        file_name = pathlib.Path(data_dir, f"{file_prefix}_n{min_size}-{max_size}.npy").resolve()
        print(f"Saving trees into {file_name}...")
        save_trees_to_file(trees, file_name, padding=max_size)
    print("Finished")


if __name__ == '__main__':
    generate_all_random_trees("../../../validation/data")



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import heapq
import itertools
import math
import multiprocessing
import operator
from collections.abc import Generator

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.tree.ser.pseudo import SubBTreePart, SubLTreePart, OPT
from slambuc.alg.util import (ipostorder_dfs, ibacktrack_chain, recreate_subtree_blocks, ipostorder_tabu_dfs,
                              ileft_right_dfs, verify_limits)


def isubtree_cutoffs(tree: nx.DiGraph, root: int = 1, lb: int = 1, ub: int = math.inf) -> tuple[tuple[int, int], int]:
    """
    Recursively return edges that cut off non-trivial subtrees from *tree* with size between *lb* and *ub*.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param lb:      lower size bound
    :param ub:      upper size bound
    :return:        cut barrier node, branches and subtree root
    """
    _tmp = {}
    for _, v in ipostorder_dfs(tree, root, inclusive=True):
        # Sum the size of descendant subtrees
        if len(brs := _tmp.setdefault(v, {s: sum(_tmp.pop(s).values(), start=1) for s in tree.successors(v)})) > 1:
            # Only consider subtrees with proper size
            for br, st in brs.items():
                if lb <= st:
                    if ub <= st:
                        _tmp[v][br] = 0
                    yield (v, br), st


def get_cpu_splits(tree: nx.DiGraph, root: int = 1, workers: int = None) -> tuple[tuple[int, int]]:
    """
    Calculate the cuts for parallelization based on *workers* count and subtree size heuristics.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param workers: workers count
    :return:        cut edges
    """
    ub = math.sqrt(len(tree) - 1)
    lb = ub if not workers else 1
    n = workers if workers else ub
    return [e for e, _ in heapq.nlargest(round(n),
                                         list(isubtree_cutoffs(tree, root, lb, ub)), key=operator.itemgetter(1))]


def isubtree_sync_cutoffs(tree: nx.DiGraph, root: int = 1,
                          size: int = math.inf) -> Generator[tuple[tuple[int], int, set[int]]]:
    """
    Recursively return edges that cut off non-trivial subtrees from *tree* with given *size*.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param size:    subtree min size
    :return:        generator of cut edge, subtree root, and related branches
    """
    _tmp, sync = {}, collections.defaultdict(set)
    for _, v in ipostorder_dfs(tree, root, inclusive=True):
        # Sum the size of descendant subtrees
        brs = _tmp.setdefault(v, {s: sum(_tmp.pop(s).values(), start=0 if s in sync else 1)
                                  for s in tree.successors(v)}).items()
        for br, st in brs:
            # Only consider subtrees with proper size
            if len(brs) > 1 and size <= st:
                # Return the cut edge, number of freely processable nodes, list of sync nodes
                yield (v, br), st, sync[br]
                # Clear nodes free to process
                _tmp[v][br] = 0
                # Cache subcase node required to be waited as a sync point
                sync[v].add(br)
            elif br in sync:
                # Mark as a dependent node and propagate sync points
                sync[v].update(sync[br])
        if v == root:
            yield (PLATFORM, root), None, sync[root]


def isubtree_splits(tree: nx.DiGraph, root: int = 1) -> Generator[tuple[tuple[int], int, set[int]]]:
    """
    Return the heuristic cutoff edges of given *tree* along with the mandatory synchronization points by assuming
    the subtree size equals *sqrt(n)*.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :return:        generator of cut edge, subtree root, and related branches
    """
    yield from ((c, sync) for c, _, sync in isubtree_sync_cutoffs(tree, root, math.ceil(math.sqrt(len(tree) - 1))))


########################################################################################################################


def _btree_partitioning(ready: multiprocessing.SimpleQueue, sync: dict[int, multiprocessing.SimpleQueue],
                        tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                        cpath: set[int] = frozenset(), delay: int = 1,
                        bidirectional: bool = True) -> None | dict[tuple[int, int], dict[int, SubBTreePart]]:
    """
    Calculates minimal-cost partitioning of a subgraph with *root* using the bottom-up tree traversal approach
    while waits for subcases at sync edges.

    This function is designed for running in a separate detached subprocess and synchronizing subresults via
    *SimpleQueue* objects as an IPC method.

    :param ready:           object for signaling the end of partitioning
    :param sync:            object regarding subtrees which results need to be waited for
    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cpath:           critical path nodes
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Allocate empty dict for local subcases combined with prior subcase results
    DP = {}
    # Iterate nodes in a bottom-up traversal order except the subtrees of the sync points
    for p, v in ipostorder_tabu_dfs(tree, root, tabu=sync):
        r_v, d_v, t_v, m_v = tree[p][v][RATE], tree[p][v][DATA], tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
        # SINGLETON: calculate the default subcase of singleton partition of node v
        cost = r_v * (d_v + t_v) + sum(vs[RATE] * vs[DATA] for vs in tree.succ[v].values())
        # Only add fetching overhead for root node and omit caching to ensure lat monotonicity of upward merging
        lat = d_v + t_v if p is PLATFORM else t_v if v in cpath else 0
        DP[v] = {(lat, lat): {m_v: SubBTreePart(cost, {v})}}
        # Bottom-up propagation for considering v's descendant subcases in rearranged order (non-sync points first)
        # -->   p   ---->   v   ----[i.]---->   b
        for b in itertools.chain(filter(lambda x: x not in sync, tree.successors(v)),
                                 filter(sync.__contains__, tree.successors(v))):
            # Waiting for dependent subprocess to be finished
            if b in sync:
                DP[b] = sync[b].get()
                sync[b].close()
            # Init empty data structure of subcase T[v,b]
            _cache = collections.defaultdict(dict)
            r_b, d_b, t_b = tree[v][b][RATE], tree[v][b][DATA], tree.nodes[b][RUNTIME]
            # Calculate possible latency/memory combinations while dropping subcases of v and b from DP
            for ((lat_v, blk_lat_v), DPv), ((lat_b, blk_lat_b), DPb) in itertools.product(DP[v].items(),
                                                                                          DP.pop(b).items()):
                for (mem_v, sub_v), (mem_b, sub_b) in itertools.product(DPv.items(), DPb.items()):
                    # Latency calculation in case v -> b edge is in cpath
                    if b in cpath:
                        r_v = tree[p][v][RATE]
                        # CUT: v -> b edge is marked as a cut (opt subcases of b have the dedicated memory value 0)
                        if mem_b == OPT:
                            # Add caching overhead of inter-block call
                            top_blk_lat = blk_lat_v + math.ceil(r_b / r_v) * d_b
                            # Add invocation delay and fetch overhead of inter-block call
                            if (lat := top_blk_lat + delay + d_b + lat_b) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                        # MERGE: v -> b edge is marked as a merge
                        else:
                            # Top block's latency comes for only the single cp node v, hence n_v = 1
                            top_blk_lat = blk_lat_v + math.ceil(r_b / r_v) * blk_lat_b
                            # Calculate new latency based on the recalculated top block's latency
                            if (lat := lat_b - blk_lat_b + top_blk_lat) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                    else:
                        # Top block's attributes remain the same, either v in cpath or not
                        top_blk_lat, lat = blk_lat_v, lat_v
                    # CUT: v -> b edge is marked as a cut
                    if mem_b == OPT:
                        # Sub-partitions are just concatenated
                        mem, barr, cost = mem_v, sub_v.barr | sub_b.barr, sub_v.cost + sub_b.cost
                    # MERGE: v -> b edge is marked as a merge
                    elif (mem := mem_v + mem_b) > M:
                        # Infeasible subcase due to exceeded memory constraint
                        continue
                    else:
                        # Top blocks of the two sub-partitions are merged together with root node v
                        barr, cost = sub_v.barr | sub_b.barr - {b}, sub_v.cost + sub_b.cost - 2 * r_b * d_b
                    # Store the min cost subcase
                    sub_lats = lat, top_blk_lat
                    # Check whether the current subcase is a better sub-solution for state (v, sub_lats, mem)
                    if mem not in _cache[sub_lats] or cost < _cache[sub_lats][mem].cost:
                        if bidirectional:
                            # Eliminate prior subcases that are dominated by the new subcase
                            for l, m in [(_l, _m) for _l in _cache if _l[0] >= lat and _l[1] >= top_blk_lat
                                         for _m in _cache[_l] if _m >= mem and _cache[_l][_m].cost >= cost]:
                                if len(_cache[l]) > 1:
                                    del _cache[l][m]
                                else:
                                    del _cache[l]
                        # Add superior subcase
                        _cache[sub_lats][mem] = SubBTreePart(cost, barr)
            # Store min subcases as C(v,i-1) for the next iteration of the propagation process
            DP[v] = _cache
            # Drop unnecessary subcases
        # Store the cost-opt subcases wrt. latency for node v encoded with memory value 0
        for sub_v in DP[v].values():
            sub_v[OPT] = min(sub_v.values(), key=operator.itemgetter(0))
    # Notify waiting process and push optimal subcases or return TDP locally for the main thread
    return ready.put(DP[root]) if ready else DP


def pseudo_mp_btree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                                 L: int = math.inf, cp_end: int = None, delay: int = 1,
                                 bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes.

    Partitioning is calculated using the bottom-up tree traversal approach.

    Arbitrary disjoint subtrees are partitioned in separate subprocesses.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    workers, events, sync = [], {}, None
    # Split tree and initiate subprocesses to calculate subcases
    try:
        for (p, v), sync in isubtree_splits(tree, root):
            # Collect event of sync points
            sync = {s: events[s] for s in sync}
            if p is not PLATFORM:
                # Add own event for signaling readiness
                ready = events.setdefault(v, multiprocessing.SimpleQueue())
                # Create and start worker process
                workers.append(multiprocessing.Process(target=_btree_partitioning, name=f"subtree_{v}",
                                                       args=(ready, sync, tree, v, M, L, cpath, delay, bidirectional),
                                                       daemon=True))
                workers[-1].start()
        # Process last/topmost subtree in the main process and get subresults locally
        DP = _btree_partitioning(None, sync, tree, root, M, L, cpath, delay, bidirectional)
    except BaseException:
        # Terminate all initiated subprocesses in case of interruption
        for w in workers:
            w.terminate()
        # Reraise exception for further handling
        raise
    finally:
        # Wait for all subprocesses to terminate for closing all used resources
        for w in workers:
            w.join(timeout=0)
            if w.is_alive():
                w.kill()
                w.join()
    # Subcases under the root node contain the feasible partitioning
    if opt_lats := min(DP[root], key=lambda _l: DP[root][_l][OPT].cost, default=None):
        opt = DP[root][opt_lats][OPT]
        return recreate_subtree_blocks(tree, opt.barr), opt.cost, opt_lats[0]
    else:
        # No feasible solution
        return INFEASIBLE


########################################################################################################################


def _ltree_partitioning(ready: multiprocessing.SimpleQueue, sync: dict[int, multiprocessing.SimpleQueue],
                        tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                        L: int = math.inf, cpath: set[int] = frozenset(), delay: int = 1,
                        bidirectional: bool = True) -> None | dict[int, dict[int, SubBTreePart]]:
    """
    Calculates minimal-cost partitioning of a subgraph with *root* using the left-right tree traversal approach
    while waits for subcases at sync edges.

    This function is designed for running in a separate detached subprocess and synchronizing subresults via
    *SimpleQueue* objects as an IPC method.

    :param ready:           object for signaling the end of partitioning
    :param sync:            object regarding subtrees which results need to be waited for
    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cpath:           critical path nodes
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Allocate empty dict for local subcases combined with prior subcase results
    TDP = {}
    # Process subtrees of T in a bottom-up traversal order
    for p, n in ipostorder_tabu_dfs(tree, root, tabu=sync):
        # Init empty data structure for optimal subcases T_n[v,b]
        DP = collections.defaultdict(lambda: collections.defaultdict(dict))
        # Store sync point of the overlapped subtree
        sentinel = None
        # Traverse subtree T_n in a left-right traversal
        #       |--(j-1.)--> [v_p]   |--(i-1.)--> [b_p]                                 | left side
        # -->  [u]  ----(j.)---->  [v]  ----(i.)---->  [b]  ----(max_k.)----> [b_k]     | right side
        for e_uvp, b_p, e_vb in ileft_right_dfs(tree, n):
            v, b = e_vb
            # Merge edge u -> v
            if b == 0:
                # Cache passed sync point
                if v in sync:
                    sentinel = sync[v]
                # SINGLETON: C[n,0] - calculate the default subcase of singleton partition of root node n
                if v == n:
                    r_n, d_n, t_n = tree[p][n][RATE], tree[p][n][DATA], tree.nodes[n][RUNTIME]
                    cost = r_n * (d_n + t_n) + sum(ns[RATE] * ns[DATA] for ns in tree.succ[n].values())
                    if n in cpath:
                        nc = next(filter(lambda c: c in cpath, tree.successors(n)), None)
                        lat = d_n + t_n + (math.ceil(tree[n][nc][RATE] / r_n) * tree[n][nc][DATA] if nc else 0)
                    else:
                        lat = 0
                    DP[e_vb][lat][tree.nodes[n][MEMORY]] = SubLTreePart(cost, 1, {n})
                # MERGE: C[v,0] - subcase of merging single node v into the n-rooted block containing node u
                else:
                    u, v_p = e_uvp
                    r_v, d_v, t_v = tree[u][v][RATE], tree[u][v][DATA], tree.nodes[v][RUNTIME]
                    # v_p = 0 => no prior sibling node, C[v,0] = C[u,j-1] + delta(u,v)
                    for lat_u in DP[e_uvp]:
                        for mem_u in DP[e_uvp][lat_u]:
                            if (mem := mem_u + tree.nodes[v][MEMORY]) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            sub_u = DP[e_uvp][lat_u][mem_u]
                            # v in cpath => u in cpath
                            if v in cpath:
                                rel_r_v = math.ceil(r_v / tree[next(tree.predecessors(u))][u][RATE])
                                n_v = rel_r_v * sub_u.mul
                                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                                w_v = math.ceil(tree[v][vc][RATE] / r_v) * tree[v][vc][DATA] if vc else 0
                                if (lat := lat_u + n_v * (t_v - d_v + w_v)) > L:
                                    # Infeasible subcase due to exceeded latency constraint
                                    continue
                            else:
                                lat, n_v = lat_u, sub_u.mul
                            cost = sub_u.cost + r_v * (t_v - d_v) + sum(vs[RATE] * vs[DATA]
                                                                        for vs in tree.succ[v].values())
                            DP[e_vb][lat][mem] = SubLTreePart(cost, n_v, sub_u.barr)
            # CUT: C[v,b]
            else:
                # b_k <- C[v_i, d(v_i)]
                b_k = collections.deque(tree.successors(b), 1).pop() if len(tree.succ[b]) else 0
                # Previously merged edge: v -> b  =>  [root ... -> u -> v -> b ...]   |   [b]  --(max_k.)--> [b_w]
                DP[e_vb] = DP[b, b_k]
                # Cut edge: v -> b  |  [root ... -> u -> v] - [b ...]   |   v --(i-1.)--> b_p
                for lat_v in DP[v, b_p]:
                    for mem_v in DP[v, b_p][lat_v]:
                        # Waiting for dependent subprocess to be finished and pull optimal subcases
                        if b not in TDP:
                            TDP.update(sentinel.get())
                            sentinel.close()
                        # assert b in TDP, f"proc{root} {b} not in {TDP.keys()}, {e_uvp=}, {b_p=}, {e_vb=}"
                        for lat_b, opt_b in TDP[b].items():
                            # b in cpath => v in cpath
                            if (lat := lat_v + delay + lat_b if b in cpath else lat_v) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                            sub_v = DP[v, b_p][lat_v][mem_v]
                            cost, barr = sub_v.cost + opt_b.cost, sub_v.barr | opt_b.barr
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or mem_v not in DP[e_vb][lat] or cost < DP[e_vb][lat][mem_v].cost:
                                if bidirectional:
                                    # Eliminate prior subcases that are dominated by the new subcase
                                    for l, m in [(_l, _m) for _l in DP[e_vb] if _l >= lat for _m in DP[e_vb][_l]
                                                 if _m >= mem_v and DP[e_vb][_l][_m].cost >= cost]:
                                        if len(DP[e_vb][l]) > 1:
                                            del DP[e_vb][l][m]
                                        else:
                                            del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][mem_v] = SubLTreePart(cost, sub_v.mul, barr)
        # Cache the best subcase for subtree T_n
        n_w = collections.deque(tree.successors(n), 1).pop() if len(tree.succ[n]) else 0
        # Store the best subcases of subtree T_n
        TDP[n] = {lat_n: min(dp.values(), key=operator.itemgetter(OPT)) for lat_n, dp in DP[n, n_w].items()}
    # Notify waiting process and push optimal subcases or return TDP locally for the main thread
    return ready.put(TDP) if ready else TDP


def pseudo_mp_ltree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                                 L: int = math.inf, cp_end: int = None, delay: int = 1,
                                 bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes.

    Partitioning is calculated using the left-right tree traversal approach.

    Arbitrary disjoint subtrees are partitioned in separate subprocesses.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    workers, events, sync = [], {}, None
    try:
        # Split tree and initiate subprocesses to calculate subcases
        for (p, v), sync in isubtree_splits(tree, root):
            # Collect event of sync points
            sync = {s: events[s] for s in sync}
            if p is not PLATFORM:
                # Add own event for signaling readiness
                ready = events.setdefault(v, multiprocessing.SimpleQueue())
                # Create and start worker process
                workers.append(multiprocessing.Process(target=_ltree_partitioning, name=f"subtree_{v}",
                                                       args=(ready, sync, tree, v, M, L, cpath, delay, bidirectional),
                                                       daemon=True))
                workers[-1].start()
        # Process last/topmost subtree in the main process and get subresults locally
        TDP = _ltree_partitioning(None, sync, tree, root, M, L, cpath, delay, bidirectional)
    except BaseException:
        # Terminate all initiated subprocesses in case of interruption
        for w in workers:
            w.terminate()
        # Reraise exception for further handling
        raise
    finally:
        # Wait for all subprocesses to terminate for closing all used resources
        for w in workers:
            w.join(timeout=0)
            if w.is_alive():
                w.kill()
                w.join()
    # Subcases under the root node contain the feasible partitioning
    if opt_lat := min(TDP[root], key=lambda _l: TDP[root][_l].cost, default=None):
        opt = TDP[root][opt_lat]
        return recreate_subtree_blocks(tree, opt.barr), opt.cost, opt_lat
    else:
        # No feasible solution
        return INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .greedy import greedy_par_tree_partitioning
from .ilp import (tree_par_cfg_partitioning, tree_par_hybrid_partitioning, tree_par_mtx_partitioning,
                  all_par_tree_mtx_partitioning)
from .pseudo import pseudo_par_ltree_partitioning
from .pseudo_mp import pseudo_par_mp_ltree_partitioning



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import math
from collections.abc import Generator

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.util import (isubtrees, ibacktrack_chain, ipowerset, path_blocks, block_cost, block_latency,
                              block_memory, block_cpu)


def ichains_exhaustive(tree: nx.DiGraph, root: int, M: int, N: int) -> Generator[list[int]]:
    """
    Calculate all combination of edge cuts and returns only if it is feasible wrt. the chain connectivity, M, and N.

    Calculation is improved compared to brute force to only start calculating cuts from c_min.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound in MB
    :param N:       upper CPU core bound
    :return:        generator of chain partitions
    """
    c_min = math.ceil(sum(nx.get_node_attributes(tree, MEMORY).values()) / M) - 1
    for cuts in ipowerset(tree.edges(range(1, len(tree))), start=c_min):
        barr = {root}.union(v for _, v in cuts)
        # Check whether the subtrees are chains and meet the memory requirement M and N
        for b, nodes in isubtrees(tree, barr):
            if max(d for _, d in tree.subgraph(nodes).out_degree) > 1:
                break
            memory, rate = zip(*[(tree.nodes[v][MEMORY], tree[u][v][RATE]) for u, v in
                                 itertools.pairwise([next(tree.predecessors(b)), *nodes])])
            if block_memory(memory, 0, len(nodes) - 1) > M or block_cpu(rate, 0, len(nodes) - 1) > N:
                break
        else:
            yield barr


def ifeasible_chains(tree: nx.DiGraph, root: int, M: int, N: int) -> Generator[list[int]]:
    """
    Calculate only feasible chain partitions and returns the one which meets the limits M and N.

    Calculation is improved compared to brute force to only calculate chain partitions based on the branching nodes.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound in MB
    :param N:       upper CPU core bound
    :return:        generator of chain partitions
    """
    branch_edges = [itertools.chain(itertools.combinations(tree.succ[b], len(tree.succ[b]) - 1),
                                    [tuple(tree.successors(b))]) for b in (v for v, d in tree.out_degree if d > 1)]
    single_edges = ipowerset([v for v in tree.nodes
                              if v is not PLATFORM and tree.degree(next(tree.predecessors(v))) == 2])
    for chain_cuts in itertools.product(*branch_edges, single_edges):
        barr = {root}.union(itertools.chain.from_iterable(chain_cuts))
        # Check whether the subtrees are chains and meet the memory requirement M and N
        for b, nodes in isubtrees(tree, barr):
            memory, rate = zip(*[(tree.nodes[v][MEMORY], tree[u][v][RATE]) for u, v in
                                 itertools.pairwise([next(tree.predecessors(b)), *nodes])])
            if block_memory(memory, 0, len(nodes) - 1) > M or block_cpu(rate, 0, len(nodes) - 1) > N:
                break
        else:
            yield barr


def greedy_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = math.inf,
                             L: int = math.inf, cp_end: int = None, delay: int = 1, unit: int = 100,
                             ichains=ifeasible_chains, only_cuts: bool = False) -> list[T_RESULTS]:
    """
    Calculates minimal-cost partitioning of a service graph(tree) by iterating over all possible cuttings.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks (in MB)
    :param N:           upper CPU core bound of the partition blocks
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:       invocation delay between blocks
    :param unit:        rounding unit for the cost calculation (default: 100 ms)
    :param ichains:     generator of chain partitions
    :param only_cuts:   return the number of cuts instead of the calculated latency
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    best_res, best_cost = [INFEASIBLE], math.inf
    # Iterates over all possible cuttings
    for barr in ichains(tree, root, M, N):
        partition = []
        sum_cost = 0
        for b, nodes in isubtrees(tree, barr):
            partition.append(nodes)
            runtime, rate = zip(*[(tree.nodes[v][RUNTIME], tree[u][v][RATE])
                                  for u, v in itertools.pairwise([next(tree.predecessors(b)), *partition[-1]])])
            sum_cost += block_cost(runtime, rate, 0, len(partition[-1]) - 1, unit)
        # Calculate blocks of critical path based on the partitioning
        cp_block = path_blocks(partition, reversed(list(ibacktrack_chain(tree, root, cp_end))))
        sum_lat = sum(block_latency([tree.nodes[v][RUNTIME] for v in blk], 0, len(blk) - 1, delay, 0, len(blk) - 1)
                      for blk in cp_block) + (len(cp_block) - 1) * delay
        partition.sort()
        if sum_lat <= L:
            if only_cuts:
                sum_lat = len(cp_block) - 1
            # Store partitioning with the same best cost for comparison
            if sum_cost == best_cost:
                best_res.append((partition, sum_cost, sum_lat))
            # Initialize new best cost partitioning
            elif sum_cost < best_cost:
                best_res, best_cost = [(partition, sum_cost, sum_lat)], sum_cost
    return best_res



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib

import networkx as nx

from slambuc.alg.service import NAME, RUNTIME
from slambuc.alg.tree.dp.meta import leaf_label_nodes, isubchains, meta_tree_partitioning
from slambuc.alg.util import ichain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import print_tree_summary, evaluate_tree_partitioning


def run_test(tree: nx.DiGraph, M: int, N: int, L: int, root: int = 1, cp_end: int = None, delay: int = 10,
             unit: int = 100):
    partition, opt_cost, opt_lat = meta_tree_partitioning(tree, root, M, N, L, cp_end, delay, unit)
    # partition = recreate_barr_blocks(tree, barr) if barr else []
    evaluate_tree_partitioning(tree, partition, opt_cost, root, cp_end, M, N, L, delay, unit)
    return partition, opt_cost, opt_lat


def test_node_labeling():
    labeled_tree = leaf_label_nodes(get_random_tree(10))
    draw_tree(labeled_tree)
    print_tree_summary(labeled_tree)


def test_chain_pruning():
    labeled_tree = leaf_label_nodes(get_random_tree(10))
    draw_tree(labeled_tree)
    for chain, m in isubchains(labeled_tree, 1):
        print(f"Subchain: {chain}, branches: {m}")


def test_cp_chain():
    labeled_tree = leaf_label_nodes(get_random_tree(10))
    draw_tree(labeled_tree)
    print(list(ichain(labeled_tree, 1, 10)))


def test_tree_partitioning():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    tree.graph[NAME] += "-meta"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=15,
                  N=2,
                  L=math.inf,
                  delay=10)
    run_test(**params)


def test_random_tree_partitioning(n: int = 10):
    tree = get_random_tree(10)
    tree.graph[NAME] += "-meta"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  N=2,
                  L=math.inf,
                  delay=10)
    run_test(**params)


def test_tree_partitioning_latency():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_latency.gml", destringizer=int)
    tree.graph[NAME] += "-meta"
    M = 6
    N = 3
    root = 1
    cp_end = 10
    delay = 10
    # No restriction
    L = math.inf
    run_test(**locals())
    # Optimal solution
    L = sum(tree.nodes[v][RUNTIME] for v in (1, 3, 8, 10)) + delay * 3
    run_test(**locals())
    # Forces to reduce blocks
    L = sum(tree.nodes[v][RUNTIME] for v in (1, 3, 8, 10)) + delay * 2
    run_test(**locals())
    # Stricter latency
    L = sum(tree.nodes[v][RUNTIME] for v in (1, 3, 8, 10)) + delay * 1
    run_test(**locals())
    # Infeasible due to M
    L = sum(tree.nodes[v][RUNTIME] for v in (1, 3, 8, 10)) + delay * 0
    run_test(**locals())


if __name__ == '__main__':
    # test_node_labeling()
    # test_chain_pruning()
    # test_cp_chain()
    test_tree_partitioning()
    # test_random_tree_partitioning()
    # test_tree_partitioning_latency()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bisect
import functools
import itertools
import random

import networkx as nx


def wrand_sample(population: list[int | float], weights: list[int], k: int = 1) -> list[int | float]:
    """
    Provide an *k*-size weighted random sample from *population* without replacement according to the given *weights*.

    See more: https://stackoverflow.com/questions/43549515/weighted-random-sample-without-replacement-in-python

    :param population:  list of items
    :param weights:     list of item weights
    :param k:           sample size (default: 1)
    :return:            sample list
    """
    # Trivial solution
    if len(population) <= k:
        return population
    # Cumulative distribution of weights
    acc_weights = list(itertools.accumulate(weights))
    w_sum, opted = acc_weights[-1], set()
    while len(opted) < k:
        idx = bisect.bisect_left(acc_weights, w_sum * random.random())
        # Overwrite already drawn item to ensure that one item is chosen only for once
        opted.add(idx)
    return list(population[idx] for idx in opted)


def generate_power_ba_graph(n: int, m: int, Alpha: float = 1.0, a: float = 0.0, root: int = 0,
                            create_using: nx.Graph = None) -> nx.Graph:
    """
    Generate Barabasi-Albert (BA) graph where the probability of choosing a vertex *v* for connecting to another node
    follows a Power law distribution as *P(v) = deg(v)^Alpha + a*.

    Thus, choosing *Alpha = 1.0* and *a = 0.0* falls back to standard BA graph generation.

    Choosing *m = 1* ensures the output to be a tree by default.

    See also: https://networkx.org/documentation/stable/_modules/networkx/generators/random_graphs.html#barabasi_albert_graph
    and the related paper: https://dl.acm.org/doi/abs/10.5555/3432601.3432616.

    :param n:               number of nodes
    :param m:               number of existing nodes (or new edges) attached to the new node in each step
    :param Alpha:           power of preferential attachment (default: 1.0)
    :param a:               attractiveness of vertices with no edges (default: 0.0)
    :param root:            initial node ID that is increased in each attachment step (default: 0)
    :param create_using:    graph type to construct (default: undirected, use `nx.DiGraph` to get a directed graph)
    :return:                created graph
    """
    graph = nx.from_edgelist(((root, i) for i in range(root + 1, root + m + 1)), create_using=create_using)
    node_max, node_miss = max(graph), n - len(graph)
    for source in range(node_max + 1, node_max + node_miss + 1):
        nodes, deg_weights = zip(*((v, d ** Alpha + a) for v, d in graph.degree))
        graph.add_edges_from(zip(itertools.repeat(source), wrand_sample(nodes, deg_weights, m)))
    return graph


generate_power_ba_tree = functools.partial(generate_power_ba_graph, m=1)  # Generate power BA trees using m=1.
